/*************************************************************************
Copyright (c) :   Darko
Project   :       Struct offsets recognizing from function declaration
**************************************************************************
PROGRAM   :       H_2_STROFF.C
**************************************************************************
Revision History :
 Version     Author          Date      Description
   V1.0      Darko       30.04.2003    creation
*************************************************************************/
/* Author's email: gondor1@bigfoot.com                                  */
/*************************************************************************
Function list:
--------------------------------------------------------------------------
External functions list:
--------------------------------------------------------------------------
Special considerations:
*************************************************************************/

// IDA does a value->bool conversion, so disable the performance warning
#pragma warning( disable : 4800 )
// IDA references an internal type as both class and struct..
#pragma warning( disable : 4099 )

#include <ida.hpp>
#include <idp.hpp>
#include <name.hpp>
#include <loader.hpp>
#include <diskio.hpp>
#include <struct.hpp>
#include <frame.hpp>
#include <auto.hpp>
#include <enum.hpp>
#include <typeinf.hpp>

#define _H_2_STROFF_C                         /* enable local data */
#include "h_2_stroff.h"
#undef _H_2_STROFF_C

const char * const gpLicense[] =
{
    "ATTENTION: Parts of the code (everything in obj_rec directory) are private ",
    "           property of Halvar Flake <halvar@blackhat.com>. If you plan to use ",
    "           that source code you should talk to Halvar first!!",
    " ",
    "ATTENTION 2: The rest of the source code (all files with 'Copyright (c) :   Darko' ",
    "             in file header) is under BSD license.",
NULL
};

#ifdef USE_OBJ_REC
#include "objrec.hpp"
#else
extern int Off_stack_current_func(ea_t curr_addr, ea_t struct_addr, tid_t struct_tid, bool bEnums);
#endif

/**********************************************************************
* Function:     my_auto_wait
* Description:
* Parameters:
* Returns:
**********************************************************************/
inline void my_auto_wait(void) { do { autoWait(); } while ( !autoIsOk() ); }

/**********************************************************************
* Function:     DeleteModifiers
* Description:
* Parameters:   char *linePtr
* Returns:
**********************************************************************/
static int DeleteModifiers(char *linePtr)
{
    static const char *Modifiers[] =
    {
        "const",
        "volatile",
        "struct",
        NULL
    };

    for (int i=0; Modifiers[i] != NULL; i++)
    {
        char *ptr;
        while ( NULL != ( ptr = strstr(linePtr, Modifiers[i]) )  )
        {
            memset(ptr, ' ', strlen(Modifiers[i]));
        }
    }
    return 0;
}

/**********************************************************************
* Function:     GetFirstToken
* Description:
* Parameters:   char *linePtr
* Returns:      int
**********************************************************************/
static int GetFirstToken(char *linePtr)
{
    int i;
    for (i=0; linePtr[i] != ' ' && linePtr[i] != '*' && linePtr[i]; )
    {
        i++;
    }
    if ( i )
    {
        linePtr[i] = 0;
    }
    return i;
}

/**********************************************************************
* Function:     GetLastToken
* Description:
* Parameters:   char *linePtr, char *name
* Returns:      int
**********************************************************************/
static int GetLastToken(char *linePtr, char *name)
{
    int i;
    name[0] = 0;
    for (i=strlen(linePtr)-1; linePtr[i] != ' ' && linePtr[i] != '*' && linePtr[i] && i > 0; )
    {
        i--;
    }
    if ( i )
    {
        sscanf(&linePtr[i+1], "%s", name);
    }
    return i;
}

/**********************************************************************
* Function:     DeleteSpaces
* Description:
* Parameters:   char *linePtr
* Returns:      int
**********************************************************************/
static int DeleteSpaces(char *linePtr)
{
    int i;
    int len = strlen(linePtr);

    for (i=0; linePtr[i] == ' ' && linePtr[i] && len > 0; )
    {
        i++;
    }
    if ( i )
    {
        /* remove leading spaces */         /* include null termination */
        memmove(&linePtr[0], &linePtr[i], (int)(len - i + 1));
    }
    return i;
}

/**********************************************************************
* Function:     AnalisePointer
* Description:
* Parameters:   func_t *func, int arg_no, char *arg_name, char *arg_dec
* Returns:      int
**********************************************************************/
static int AnalisePointer(func_t *func, int arg_no, char *arg_name, char *arg_dec)
{
    tid_t struct_id = BADNODE ;

    if ( arg_dec[0] )
    {
        struct_id = get_struc_id(arg_dec);
        if ( struct_id == BADNODE )
        {
            /* maybe type is in TIL database, so move it to IDB */
            til2idb(-1, arg_dec);
            /* give it one more try */
            struct_id = get_struc_id(arg_dec);
        }
        struc_t *fs = get_frame(func);

        //if ( struct_id == BADNODE || fs == NULL )
        //{
        //    msg("processing '%s' '%s' %X %p failed!\n", arg_name, arg_dec, struct_id, fs);
        //}

        if ( arg_name[0] && struct_id != BADNODE && fs != NULL )
        {
            ulong rel_offset;
            member_t *mem =  get_member_by_name(fs, arg_name);
            if ( mem != NULL )
            {
                rel_offset = mem->soff - func->frsize;
                //{
                //    char bufn[100];
                //    get_func_name(func->endEA, bufn, sizeof(bufn));
                //
                //    msg("processing member %s at offset %X for '%s' at %X\n",
                //        arg_name, rel_offset, bufn, func->startEA);
                //}
#ifdef USE_OBJ_REC
                ReconstructFromAddr(func->startEA, arg_name, rec_full, arg_dec);
#else
                Off_stack_current_func(func->startEA, rel_offset, struct_id, false);
#endif
            }
        }

        if ( arg_name[0] == 0 && struct_id != BADNODE && fs != NULL )
        {
            /* argument has data type but not name */
            ulong rel_offset = 0;
            // msg("no name for argument!\n");

            member_t *mem = NULL;

            /* if first arg is not named 'arg_0' then we are in trouble :) */
            mem =  get_member_by_name(fs, "arg_0");
            if ( mem != NULL )
            {
                rel_offset = (mem->eoff - mem->soff);
            }
            else
            {
                rel_offset = 4;         /* last attempt */
            }

            for (int cnt=0; cnt < arg_no ; cnt++)
            {
                mem = get_best_fit_member(fs, rel_offset);
                if ( mem == NULL )
                {
                    break;
                }
                rel_offset += (mem->eoff - mem->soff);
                //msg("NEXT rel_offset %X argno %d\n", rel_offset, arg_no);
            }

            if ( mem != NULL )
            {
                rel_offset = mem->soff - func->frsize;
                //msg("processing member %s at offset %X for %X\n",
                //        get_member_fullname(mem->id),
                //        rel_offset, func->startEA);
                //msg("name = %s\n", get_member_name(mem->id));
#ifdef USE_OBJ_REC
                ReconstructFromAddr(func->startEA, get_member_name(mem->id), rec_full, arg_dec);
#else
                Off_stack_current_func(func->startEA, rel_offset, struct_id, false);
#endif
            }
        }
    }
    return 0;
}

/**********************************************************************
* Function:     ParseArgument
* Description:
* Parameters:   char *arg_dec, func_t *func, int arg_no
* Returns:      int
**********************************************************************/
static int ParseArgument(char *arg_dec, func_t *func, int arg_no)
{
    char arg_name[MAXNAMELEN + 1];
    char *ptr      = strrchr(arg_dec, '*');
    char *ptr_last = strchr(arg_dec, '*');
    if ( ptr != NULL && (ptr == ptr_last) )
    {
        DeleteModifiers(arg_dec);
        DeleteSpaces(arg_dec);
        GetLastToken(arg_dec, arg_name);
        GetFirstToken(arg_dec);
        //msg("arg%d '%s' is a pointer to '%s'\n", arg_no, arg_name, arg_dec);
        AnalisePointer(func, arg_no, arg_name, arg_dec);
    }
    return 0;
}


/**********************************************************************
* Function:     ProcessDeclaration
* Description:
* Parameters:   char *header, func_t *func
* Returns:      int
**********************************************************************/
static int ProcessDeclaration(char *header, func_t *func )
{
    char arg_dec[MAXNAMELEN + 1];
    char *ptr = strrchr(header, '(');
    int len = strlen(header);
    if ( ptr != NULL && (len + header) > ptr )
    {
        int len_to_end = strlen(ptr);
        memmove(header, ptr+1, len_to_end);
        ptr = strrchr(header, ')');
        if ( ptr != NULL )
        {
            *ptr = 0;                   /* loose ending ')' */
        }

        //msg("function: '%s'\n", header);
        int arg_no = 1;
        do
        {
            ptr = strchr(header, ',');
            memset(arg_dec, 0, sizeof(arg_dec));
            if ( ptr != NULL )
            {
                len = (ulong)ptr - (ulong)header;
                memcpy(arg_dec, header, len);
                arg_dec[len] = 0;
                memmove(header, ptr+1, strlen(header) - len);
            }
            else
            {
                strcpy(arg_dec, header);
                header[0] = 0;
            }
            ParseArgument(arg_dec, func, arg_no);
            arg_no++;
        } while ( ptr != NULL );
    }
    return 0;
}


/**********************************************************************
* Function:    PluginMain
* Description: This function contains the core plugin code.
* Parameters:  int arg
* Returns:     none
**********************************************************************/
void idaapi PluginMain( int arg )
{
    my_auto_wait();                     /* wait until IDA is ready */

    for ( int cnt=0; cnt < get_func_qty(); cnt++)
    {
        char buff[MAXNAMELEN + 1];
        func_t *curr_func = getn_func(cnt);

        if ( curr_func != NULL )
        {
            bool bOK = print_type(curr_func->startEA, buff, sizeof(buff)-1, true);

            if ( bOK )
            {
                ProcessDeclaration(buff, curr_func);
                update_func(curr_func);     /* done, update IDA database */
            }
        }
    }

    my_auto_wait();                     /* wait until IDA is ready */

    return ;
}


/**********************************************************************
* Function:     PluginStartup
* Description:
*   IDA will call this function only once. If this function returns PLUGIN_SKIP,
*   IDA will never load it again. If it returns PLUGIN_OK, IDA will unload the plugin
*   but remember that the plugin agreed to work with the database. The plugin will
*   be loaded again if the user invokes it by pressing the hotkey or selecting it
*   from the menu. After the second load, the plugin will stay in memory.
* Parameters:  none
* Returns:     none
**********************************************************************/
int idaapi  PluginStartup(void)
{
    // We only support running our plugin when a PE file is loaded..
    if ( inf.filetype != f_PE )
    {
        return PLUGIN_SKIP;
    }
    if ( strcmp(inf.procName, "metapc") != 0 )
    {
        msg(" CPU Type should be \"metapc\" but is %s ! Aborting ...\n", inf.procName);
        return PLUGIN_SKIP;
    }
    return PLUGIN_OK;
}

/**********************************************************************
* Function:    PluginShutdown
* Description: IDA will call this function when the user asks to exit.
*              This function is *not* called in the case of emergency exits.
* Parameters:  none
* Returns:     none
**********************************************************************/
void idaapi PluginShutdown(void)
{
    return ;
}

/*******************************************************************************
*
*   Strings required for IDA Pro's PLUGIN descriptor block
*
********************************************************************************/
char    gacPluginComment[]   = "Struct offsets recognizing based on arguments from function declaration";
char    gacPluginShortName[] = "H_2_OFF    V0.4";
char    gacPluginHelpText[]  = "H_2_OFF\n\n"
                               "This module allows you to propagate struct offsets "
                               "from function declaration\n";


/*******************************************************************************
*
*   This 'PLUGIN' data block is how IDA Pro interfaces with this plugin.
*
********************************************************************************/
extern "C" plugin_t PLUGIN =
{
    IDP_INTERFACE_VERSION,      // Version
    PLUGIN_UNL,                 // Plugin Flags: unload the plugin immediately after run
    PluginStartup,              // Plugin Startup
    PluginShutdown,             // Plugin Shutdown.. this can be NULL
    PluginMain,                 // Plugin Main

    gacPluginComment,           // Comment about the plugin
    gacPluginHelpText,          // Multiline help about the plugin

    gacPluginShortName,         // Preferred short name of the plugin
    "",                         // Preferred hotkey to run the plugin (no hotkey specified here)
};

