/*
 * This plugin converts all indirect ptrs to struct member to explicit
 * offset inside struct
 * Version 0.1, 24-IV-2001, (C) RedPlait, redplait@usa.net
 * Version 0.2  18.04.2003 14:54, Darko <gondor1@bigfoot.com>
 */

#pragma warning( disable : 4800 )
#pragma warning( disable : 4099 )


#include <ida.hpp>
#include <idp.hpp>
#include <loader.hpp>
#include "rp.hpp"
#include <funcs.hpp>
#include <struct.hpp>
#include <search.hpp>
#include "regs.hpp"

/***** for debugging *****/
#include "rp_debug.h"

/* get start addr, end addr and do all dirty work */
ea_t work_horse(ea_t start, ea_t end, ea_t struct_addr, tid_t struct_tid)
{
#ifdef RP_DEBUG
    msg("work_horse(%X, %X)\n", start, end);
#endif

    my_vm   vm;
    ulong size;

    while ( start < end )
    {
        size = vm.analize_instr(start, struct_addr, struct_tid);
        if ( size < 1 )
        {
            return BADADDR;
        }
        start += size;
        if ( !vm.actual )
        {
            return start;
        }
    }
    return start;
}

/* get start addr, end addr and do all dirty work */
ea_t work_horse_enum(ea_t start, ea_t end, ea_t struct_addr, tid_t struct_tid)
{
#ifdef RP_DEBUG
    msg("work_horse_enum(%X, %X)\n", start, end);
#endif

    my_vm   vm;
    ulong size;

    while ( start < end )
    {
        size = vm.analize_instr_enum(start, struct_addr, struct_tid);
        if ( size < 1 )
        {
            return BADADDR;
        }
        start += size;
        if ( !vm.actual )
        {
            return start;
        }
    }
    return start;
}


/* processed given stack struct in indicated function */
int Off_stack_current_func(ea_t curr_addr, ea_t struct_addr, tid_t struct_tid, bool bEnums)
{
    int res = 0;
    ea_t from;

    func_t *curr_func = get_func(curr_addr);

    if ( NULL == curr_func )
    {
        msg("Address is not in a function (%X)\n", curr_addr);
        return 0;
    }

#ifdef RP_DEBUG
    msg("Adr str %X  fun %X offset %X\n", struct_addr, curr_func->startEA, curr_addr-struct_addr);
    msg("From %X\n", find_imm(from = curr_func->startEA-1, SEARCH_DOWN | SEARCH_NEXT, struct_addr, NULL));
#endif

    for ( from = curr_func->startEA-1; (from != BADADDR) && (from < curr_func->endEA); )
    {
        from = find_imm(from, SEARCH_DOWN | SEARCH_NEXT, struct_addr, NULL);

        if ( from < curr_func->startEA && from != BADADDR )
        {
#ifdef RP_DEBUG
            msg("bug curr_func stack %X\n", from);
#endif
            return res;
        }

        if ( bEnums )
        {
            from = work_horse_enum(from, curr_func->endEA, struct_addr, struct_tid);
        }
        else
        {
            from = work_horse(from, curr_func->endEA, struct_addr, struct_tid);
        }
        res++;
    }
    return res;
}

/* processed given struct in indicated function */
int Off_current_func(ea_t curr_addr, ea_t struct_addr, tid_t struct_tid, bool bEnums)
{
    int res = 0;
    ea_t from, addr;

    func_t *curr_func = get_func(curr_addr);

    if ( NULL == curr_func )
    {
        msg("Address is not in a function (%X)\n", curr_addr);
        return 0;
    }

    addr = struct_addr;
#ifdef RP_DEBUG
    msg("Adr %X\n", addr);
    msg("From %X\n", get_first_dref_to(addr));
#endif
    for ( from = get_first_dref_to(addr);
          from != BADADDR && from < curr_func->endEA;
          from = get_next_dref_to(addr, from) )
    {
#ifdef RP_DEBUG
        msg("from %X\n", from);
#endif
        if ( from < curr_func->startEA )
        {
            continue;
        }
        if ( bEnums )
        {
            work_horse_enum(from, curr_func->endEA, struct_addr, struct_tid);
        }
        else
        {
            work_horse(from, curr_func->endEA, struct_addr, struct_tid);
        }
        res++;
    }
    return res;
}


int Off_all_funcs(ea_t struct_addr, tid_t struct_tid, bool bEnums)
{
    int res = 0;
    func_t *f;
    ea_t adr, from;

    adr = struct_addr;
    for ( from = get_first_dref_to(adr);
          from != BADADDR;
          from = get_next_dref_to(adr, from) )
    {
        f = get_func(from);
        if ( NULL == f )
        {
          msg("Address %X is not in function\n", from);
          continue;
        }
        if ( bEnums )
        {
            work_horse_enum(from, f->endEA, struct_addr, struct_tid);
        }
        else
        {
            work_horse(from, f->endEA, struct_addr, struct_tid);
        }
        res++;
    }
    return res;
}

