#pragma warning( disable : 4800 )
#pragma warning( disable : 4099 )

#include "regs.hpp"
#include "x86.hpp"
#include "rp.hpp"
#include <bytes.hpp>
#include <ua.hpp>

#ifdef RP_DEBUG
# include "pic.hpp"
#endif

/* for debugging */
#include "rp_debug.h"

int my_vm::get_hreg(int hr)
{
    switch (hr)
    {
        case r_AL:
        case r_AH: return r_EAX;

        case r_CL:
        case r_CH: return r_ECX;

        case r_DL:
        case r_DH: return r_EDX;

        case r_BL:
        case r_BH: return r_EBX;

        default:
            return -1;
    }
    return -1;
}

void my_vm::reset_regs()
{
    actual = 0;
    for ( int i = 0; i < 8; i++ )
    {
        regs[i].reset();
    }
}

void my_vm::reset_reg(int r)
{
    if ( r > 7 )
    {
        return;
    }
    if ( regs[r].is_s )
    {
        actual--;
        regs[r].reset();
    }
}

void my_vm::set_reg(int r, tid_t tid)
{
    if ( r > 7 )
    {
        return;
    }
    if ( !regs[r].is_s )
    {
        actual++;
    }
    regs[r].set_s(tid);
}

void my_vm::xfer_reg(int from, int to)
{
    if ( regs[to].is_s )
    {
        reset_reg(to);
    }
    if ( !regs[from].is_s )
    {
        return;
    }
    set_reg(to, regs[from].tid);
    regs[to].offset = regs[from].offset;
}

static int is_end_instr(int code)
{
    switch (code)
    {
        case Ix86_call0:
        case Ix86_call1:
        case Ix86_call2:
        case Ix86_jmp0:
        case Ix86_jmp1:
        case Ix86_jmp2:
        case Ix86_jmp3:
        case Ix86_ret:
        case Ix86_retf:
            return 1;

        default:
            return 0;
    }
    return 0;
}

/* Something similar to VM
 * Returns: size of analyzed instruction
 * Logic: unpredictable :-)
 */
int my_vm::analize_instr(ea_t addr, ea_t struct_addr, tid_t struct_tid)
{
    int size = ua_ana0(addr);
    if ( size < 1 )
    {
        return size;
    }

   /* MONITOR_FUNCTION_PARSING(0x401000, 0x0040101F); */

#ifdef RP_DEBUG
    msg("analize: %X, size %d stadr %X\n", addr, size, struct_addr);
#endif
    /* check for mov reg, ds:[offset to struct] instr */
    bool mem_to_reg   = (cmd.Op1.type == o_reg && cmd.Op2.type == o_mem);
    /* Darko, 19.04.2003 14:06, check for mov reg, [ebp - stackvar] instr */
    bool stack_to_reg = (cmd.Op1.type == o_reg && cmd.Op2.type == o_displ &&
                         exists_addr(&cmd.Op2) && !exists_value(&cmd.Op2) );
    if ( cmd.itype == Ix86_mov && ( mem_to_reg || stack_to_reg ) &&
         cmd.Op1.reg < 7 &&
         cmd.Op1.dtyp == dt_dword && cmd.Op2.dtyp == dt_dword )
    {
        ea_t op_addr = cmd.Op2.addr;
        if ( struct_addr == op_addr )
        {
#ifdef RP_DEBUG
            msg("Track reg %d\n", cmd.Op1.reg);
#endif
            set_reg(cmd.Op1.reg, struct_tid);
            return size;
        }
        reset_reg(cmd.Op1.reg);
        if ( regs[cmd.Op2.reg].is_s && cmd.Op1.type == o_reg && cmd.Op2.type == o_displ )
        {
            /* check if mov reg, [reg2 + off] instruction when reg2 is traced */
            /* there is still cmd.Op2 to process, so do not return now */
        }
        else
        {
            return size;
        }
    }

    /* Darko, 19.04.2003 21:26 */
    bool reg_to_reg =
                (cmd.Op1.type == o_reg && cmd.Op2.type == o_reg) &&
                (cmd.Op1.reg < 7 && cmd.Op2.reg < 7) &&
                (cmd.Op1.dtyp == dt_dword && cmd.Op2.dtyp == dt_dword );

    /* check for interregister moving  mov */
    if ( cmd.itype == Ix86_mov && reg_to_reg )
    {
#ifdef RP_DEBUG
        msg("transfer from %d -> %d\n", cmd.Op2.reg, cmd.Op1.reg);
#endif
        xfer_reg(cmd.Op2.reg, cmd.Op1.reg);
        return size;
    }

    /* check for (sadly) base reg modification (in present time I can track only
     * one instruction: add reg, const (const < struct_size !)
     */

    if ( cmd.itype == Ix86_add &&
         cmd.Op1.type == o_reg &&
         cmd.Op2.type == o_imm &&
         cmd.Op1.reg < 7 )
    {
#ifdef RP_DEBUG
        msg("Add to reg\n");
#endif
        if ( regs[cmd.Op1.reg].is_s )
        {
#ifdef RP_DEBUG
            msg("Offset is %d\n", cmd.Op2.value);
#endif
            regs[cmd.Op1.reg].offset += cmd.Op2.value;
            return size;
        }
        return size;
    }
    /* check for something usefull :-) We have two cases:
     * 1) mov something, [reg+off]
     * 1a) mov something, [reg]
     * 2) mov[reg+off], something
     * 2a) mov[reg], something
     */
    if (  ( (cmd.Op1.type == o_displ && exists_addr(&cmd.Op1)) ||
            (cmd.Op1.type == o_phrase)    /* Darko, 18.04.2003 14:53 */
          )
         &&
        !exists_value(&cmd.Op1) &&
        cmd.Op1.phrase < 7      &&
        regs[cmd.Op1.phrase].is_s )
    {
        tid_t tid = regs[cmd.Op1.phrase].tid;
        //    rp_set_comment(addr, get_struc_name(tid), true);
        struc_t *s = get_struc(tid);
        op_stroff(addr, 0, &s->id, sizeof(tid_t), regs[cmd.Op1.phrase].offset);
    }
    /* second case */
    if (  ( (cmd.Op2.type == o_displ && exists_addr(&cmd.Op2)) ||
            (cmd.Op2.type == o_phrase )   /* Darko, 18.04.2003 14:54 */
          )
        &&
        !exists_value(&cmd.Op2) &&
        cmd.Op2.phrase < 7      &&
        regs[cmd.Op2.phrase].is_s )
    {
        tid_t tid = regs[cmd.Op2.phrase].tid;
        //    rp_set_comment(addr, get_struc_name(tid), true);
        struc_t *s = get_struc(tid);
        op_stroff(addr, 1, &s->id, sizeof(tid_t), regs[cmd.Op2.phrase].offset);
    }
  /* also we consider all calls as killed all registers */
    if ( is_end_instr(cmd.itype) )
    {
        reset_regs();
        return size;
    }
    /* O`k, next we need to check if our reg is trashed */
    const struct RP_opcode *rp_op = get_opcode(&cmd);
    if ( NULL != rp_op )
    {
        for ( int i = 0; i < 7; i++ )
        {
            if ( is_change_reg(&cmd, rp_op, i) )
            {
                reset_reg(i);
            }
        }
    }
    return size;
}


int my_vm::analize_instr_enum(ea_t addr, ea_t enum_addr, tid_t enum_tid)
{
    int size = ua_ana0(addr);
    if ( size < 1 )
    {
        return size;
    }

    //MONITOR_FUNCTION_PARSING(0x401480, 0x0040169C);

#ifdef RP_DEBUG
    msg("analize: %X, size %d stadr %X\n", addr, size, enum_addr);
    //report_instr(addr);
#endif
    /* check for mov mem, #imm instr */
    bool imm_to_mem   = (cmd.Op1.type == o_mem && cmd.Op2.type == o_imm);
    /* Darko, 19.04.2003 14:06, check for mov [ebp - stackvar], #imm instr */
    bool imm_to_stack = (cmd.Op2.type == o_imm && cmd.Op1.type == o_displ &&
                         exists_addr(&cmd.Op1) && !exists_value(&cmd.Op1) );

    if ( cmd.itype == Ix86_mov && ( imm_to_mem || imm_to_stack ) &&
         cmd.Op1.reg < 7 &&
         cmd.Op1.dtyp == dt_dword && cmd.Op2.dtyp == dt_dword )
    {
        ea_t op_addr = cmd.Op1.addr;
        if ( enum_addr == op_addr )
        {
#ifdef RP_DEBUG
            msg("ENUM \n");
#endif
            op_enum(addr, 1, enum_tid, 0);
        }
    }
    return size;
}
