#ifndef  __ANALYS
#define  __ANALYS

// the code analyser part

// IDA does a value->bool conversion, so disable the performance warning
#pragma warning( disable : 4800 )
// IDA references an internal type as both class and struct..
#pragma warning( disable : 4099 )

#include <ida.hpp>
#include <idp.hpp>
#include <name.hpp>
#include <loader.hpp>
#include <diskio.hpp>
#include <struct.hpp>
#include <frame.hpp>
#include <enum.hpp>

#include "legacy.hpp"
#include "graphnode.hpp"
#include "ptrgraph.hpp"
#include "analys.hpp"

//--------------------------------------------------------------------------
//      The code analyser section ... woo

/* Darko, 02.05.2003 18:01 - start */
void  CreateStructFromGraph(PtrGraph *grph, char *szStrucName)
{
   unsigned int   num_nodes, i;
   struc_t        *structure;

   tid_t tmp_id = get_struc_id(szStrucName);

   if ( BADNODE != tmp_id )
   {
       structure = get_struc(tmp_id);   // if it already exists, use this one
       //msg("exists %s\n", szStrucName);
   }
   else
   {
       structure = get_struc(add_struc(BADADDR, szStrucName, false));
       //msg("new structure %s\n", szStrucName);
   }

   if (structure == 0)
   {
        msg("error in CreateStructFromGraph for struct name '%s'\n", szStrucName);
        return ;
   }

   num_nodes = grph->GetNumNodes(); // number of nodes to cycle through
   for(i = 1; i < num_nodes; i++)   // cycle through all nodes
   {
      if(grph->FindNodeByIndex(i)->GetReg() == r_none)
         AnalyseNodeStackVar(grph->FindNodeByIndex(i), structure);
      else
         AnalyseNodeRegister(grph->FindNodeByIndex(i), structure);
   }
   save_struc(structure);               /* Darko, 02.05.2003 17:29 */
}
/* Darko, 02.05.2003 18:01 - end */

void  AnalyseNodeStackVar(GraphNode *node, struc_t *struc)
{
      // hrmm... do I need to do anything here ?
      return;
}

/* Darko, 29.04.2003 22:08 - start */
static int GetSize(int data_type)
{
    //msg("size %d", data_type);
    if ( data_type >= dt_byte && data_type <= dt_double  )
    {
        return (data_type * 2);
    }
    return 0;
}

void  AnalyseNodeRegister(GraphNode *node, struc_t *struc)
{
   ea_t  begin, end, curr;
   int   iOpNum;
   reg_t reg;
   insn_t   cmd_A;

   reg = node->GetReg();
   begin = node->GetBegin();
   end = node->GetEnd();
   curr = begin;

   while(curr <= end)
   {
      curr = register_xs(reg, curr, end);
                /* Darko, 01.05.2003 22:55 */
      if(curr == BADADDR)   // no further occurence found
         return;
//      msg("analysing %lx\n", curr);

      int scmd = ua_ana0(curr);     // analyze the instruction
      if ( scmd < 1 )
      {
         msg("ua_ana0 failed at %X", curr);
         curr++;
         continue;
      }
      memcpy(&cmd_A, &cmd, sizeof(insn_t));

      iOpNum = reg_opnd_num(reg, curr);

      if(iOpNum == -1)
         msg("[!] register_xs fucked up at %lx\n", curr);

      if((cmd_A.itype == x86_movA)||    // I am not sure if I got them all
         (cmd_A.itype == x86_movB)||
         (cmd_A.itype == x86_movsx)||
         (cmd_A.itype == x86_movzx)||
         (cmd_A.itype == x86_lea)||
         (cmd_A.itype == x86_or)||
         (cmd_A.itype == x86_xor)||
         (cmd_A.itype == x86_and)||
         (cmd_A.itype == x86_not)||
         (cmd_A.itype == x86_mul)||
         (cmd_A.itype == x86_imul)||
         (cmd_A.itype == x86_neg)||
         (cmd_A.itype == x86_ror)||
         (cmd_A.itype == x86_rol)||
         (cmd_A.itype == x86_shl)||
         (cmd_A.itype == x86_shr)||
         (cmd_A.itype == x86_cmp)||
         (cmd_A.itype == x86_test)||
         (cmd_A.itype == x86_inc)||
         (cmd_A.itype == x86_dec))
      {
         if(cmd_A.Operands[iOpNum].type != o_reg)
         {
            if(cmd_A.Operands[iOpNum].addr != 0)
               op_to_struc_member(struc, cmd_A.Operands[iOpNum].addr, cmd_A.ea, iOpNum, GetSize(cmd_A.Operands[iOpNum].dtyp));
            else if (cmd_A.Operands[iOpNum].value != 0)
               op_to_struc_member(struc, cmd_A.Operands[iOpNum].value, cmd_A.ea, iOpNum, GetSize(cmd_A.Operands[iOpNum].dtyp));
            else
               op_to_struc_member(struc, 0, cmd_A.ea, iOpNum, GetSize(cmd_A.Operands[iOpNum].dtyp));
         }
         // END of parsing mov* & lea ... seems buggy
      }
      else if((cmd_A.itype == x86_add)||(cmd_A.itype == x86_sub))
      {
#if !defined(H_2_STROFF)                /* Darko,30.04.2003 17:58 */
         msg("Analysing sub/add at %lx\n", cmd_A.ea);
         msg("iOpNum == %lx\n", iOpNum);
#endif

         if((iOpNum == 0)&&(cmd_A.Op1.type == o_reg))
            op_to_struc_member(struc, cmd_A.Op2.value, cmd_A.ea, 1, cmd_A.Op2.dtyp);

         if(cmd_A.Operands[iOpNum].type != o_reg)
         {
//            msg("cmd_A.Operands[iOpNum].type != o_reg\n");
//            msg("cmd_A.Operands[iOpNum].value = %lx\n", cmd_A.Operands[iOpNum].value);
//            msg("cmd_A.Operands[iOpNum].addr = %lx\n", cmd_A.Operands[iOpNum].addr);

            if(cmd_A.Operands[iOpNum].value != 0)
               op_to_struc_member(struc, cmd_A.Operands[iOpNum].value, cmd_A.ea, iOpNum, GetSize(cmd_A.Operands[iOpNum].dtyp));
            else if(cmd_A.Operands[iOpNum].addr != 0)
               op_to_struc_member(struc, cmd_A.Operands[iOpNum].addr, cmd_A.ea, iOpNum, GetSize(cmd_A.Operands[iOpNum].dtyp));
         }
      }
      else if((cmd_A.itype == x86_push)&&(cmd_A.Op1.type != o_reg))
         op_to_struc_member(struc, cmd_A.Op1.addr, cmd_A.ea, 0, GetSize(cmd_A.Op1.dtyp));

      if(cmd_A.size == 0)
         curr++;
      else
         curr += cmd_A.size;
   }  // end of while loop

}
/* Darko, 29.04.2003 22:08 - end */


#endif
