#ifndef  __GRAPHNODE
#define  __GRAPHNODE

#include <stdio.h>
#include <string.h>

// IDA does a value->bool conversion, so disable the performance warning
#pragma warning( disable : 4800 )
// IDA references an internal type as both class and struct..
#pragma warning( disable : 4099 )

#include <ida.hpp>
#include <idp.hpp>
#include <name.hpp>
#include <loader.hpp>
#include <diskio.hpp>
#include <struct.hpp>
#include <enum.hpp>

#include "insx86.hpp"
#include "graphnode.hpp"

const char *GraphNode::GetRegString(const reg_t reg)
{
   static const char * const name[] = {"eax", "ecx", "edx", "ebx", "esp", "ebp", "esi", "edi", "none"};

   return (name[reg]);
}

GraphNode *GraphNode::GetDownLink(const unsigned int idx)
{
    if ( idx <= vec_link_down.size() && idx )
    {
        return vec_link_down[idx-1];
    }
    return NULL;
}

GraphNode *GraphNode::GetUpLink(const unsigned int idx)
{
    if ( idx <= vec_link_up.size() && idx )
    {
        return vec_link_up[idx-1];
    }
    return NULL;
}

GraphNode::GraphNode()
{
   vec_link_down.reserve(100);
   vec_link_down.clear();

   vec_link_up.reserve(100);
   vec_link_up.clear();

   node_name = 0;
   node_text = 0;
   ea_begin = 0;
   ea_end = 0;
   stackvar_ofs = 0;
   reg_ptr = r_none;
   vert_order = BADNODE;
}

void GraphNode::SetNodeName(const char *nuname)
{
   if (node_name != 0)
   {
      qfree(node_name);
   }
   node_name = (char *)qalloc(strlen(nuname)+1);
   strncpy(node_name, nuname, strlen(nuname)+1);
}

void GraphNode::SetNodeText(const char *text)
{
   if (node_text != 0)
   {
      qfree(node_text);
   }
   node_text = (char *)qalloc(strlen(text)+1);
   strncpy(node_text, text, strlen(text)+1);
}

GraphNode::~GraphNode()
{
   if (node_name != 0)
   {
      qfree(node_name);
      node_name = 0;
   }

   if (node_text != 0)
   {
      qfree(node_text);
      node_text = 0;
   }
}

// This function is going to be painfully slow due to the qrealloc
// upon every node addition and due to unnecessare calling upwards & downwards
// in the graph...ahwell, fuckit. I will improve performance
// at a later point.
// These functions should never ever return false (!), something insanely
// wrong must've happened if they do.

bool GraphNode::AddDownLink(class GraphNode *child)
{
    for (unsigned int idx=0; idx < vec_link_down.size(); idx++ )
    {
        if ( vec_link_down[idx] == child )
        {
            return true;
        }
    }

    if ( vec_link_down.size() + 5 > vec_link_down.capacity() )
    {
        if ( vec_link_down.size() + 100 > vec_link_down.max_size() )
        {
            return false;
        }
        vec_link_down.reserve(vec_link_down.size() + 100);
    }

    vec_link_down.push_back(child);

    return ( child->AddUpLink(this) );
}

bool GraphNode::AddUpLink(class GraphNode *parent)
{
    for (unsigned int idx=0; idx < vec_link_up.size(); idx++ )
    {
        if ( vec_link_up[idx] == parent )
        {
            return true;
        }
    }

    if ( vec_link_up.size() + 5 > vec_link_up.capacity() )
    {
        if ( vec_link_up.size() + 100 > vec_link_up.max_size() )
        {
            return false;
        }
        vec_link_up.reserve(vec_link_up.size() + 100);
    }

    vec_link_up.push_back(parent);

    return ( parent->AddDownLink(this) );
}
#endif
