#ifndef  __GRAPHNODE_H
#define  __GRAPHNODE_H

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

// IDA does a value->bool conversion, so disable the performance warning
#pragma warning( disable : 4800 )
// IDA references an internal type as both class and struct..
#pragma warning( disable : 4099 )

#include <ida.hpp>
#include <idp.hpp>
#include <name.hpp>
#include <loader.hpp>
#include <diskio.hpp>
#include <struct.hpp>
#include <frame.hpp>
#include <auto.hpp>
#include <enum.hpp>
#include <typeinf.hpp>

#include <vector>
#include <list>
#include <map>

#include "insx86.hpp"

class GraphNode
{
private:
   std::vector<GraphNode *>         vec_link_down; // pointer to an array of "children" links
   std::vector<GraphNode *>         vec_link_up;   // pointer to an array of "parent"

   char              *node_name;    // pointer to the node name
   char              *node_text;    // additional text to fill into
                                    // the node
   unsigned int      ea_begin;      // begin of code block
   unsigned int      ea_end;        // end of code block
   reg_t             reg_ptr;       // which register (r_none if stackvar)
   unsigned int      stackvar_ofs;  // offset of the traced stackvar in
                                    // the frame
   unsigned int      vert_order;     // Vertical level for graph formatting
public:
   GraphNode();   // constructor
   ~GraphNode();  // destructor
   unsigned int  GetLinkNumUp(){return vec_link_up.size();};   // Get Number of uplinks
   unsigned int  GetLinkNumDown(){return vec_link_down.size();};  // Get Number of downlinks
   bool  AddDownLink(class GraphNode *child);
   bool  AddUpLink(class GraphNode *parent);
   void  SetNodeName(const char *nuname);
   void  SetNodeText(const char *text);
   inline void  SetBegin(unsigned int begin){ea_begin = begin;}
   inline void  SetEnd(unsigned int end){ea_end = end;}
   inline void  SetReg(reg_t reg){reg_ptr = reg;}
   inline void  SetVarOfs(unsigned int ofs){stackvar_ofs = ofs;}
   inline void  SetVertical(unsigned int order){vert_order = order;}

   inline reg_t GetReg(){return reg_ptr;}
   const char  *GetRegString(const reg_t reg);
   inline unsigned int GetVertical(){return vert_order;}
   inline unsigned int GetVarOfs(){return stackvar_ofs;}
   inline unsigned int GetBegin(){return ea_begin;}
   inline unsigned int GetEnd(){return ea_end;}
   inline const char *GetNodeText(){return node_text;}
   inline const char *GetNodeName(){return node_name;}
   GraphNode *GetDownLink(const unsigned int idx);
   GraphNode *GetUpLink(const unsigned int idx);
};

#endif
