#ifndef  __LEGACY
#define  __LEGACY

// These are old functions who've been written for the first or the second
// incarnation of the structure reconstructor which I am still using in
// the code analyser or in the pushcall recursion module

// IDA does a value->bool conversion, so disable the performance warning
#pragma warning( disable : 4800 )
// IDA references an internal type as both class and struct..
#pragma warning( disable : 4099 )

#include <ida.hpp>
#include <idp.hpp>
#include <name.hpp>
#include <loader.hpp>
#include <diskio.hpp>
#include <struct.hpp>
#include <frame.hpp>
#include <enum.hpp>

#include "insx86.hpp"
#include "objrec.hpp"
#include "legacy.hpp"

// Creates a structure member
/* Darko, 29.04.2003 21:53 - start */
bool    __stdcall op_to_struc_member(struc_t *structure, unsigned int offset, ea_t ea, int opnum, int size)
{
   char  sz_memname[1000];

   snprintf(sz_memname, sizeof(sz_memname)-1, "%s_%lx", StrucName, offset);
   if(offset > 0x8000000)
      msg("Adding structure member %s at offset %lx ! %lx\n", sz_memname, offset, ea);

#if !defined(H_2_STROFF)                /* Darko,30.04.2003 17:58 */
   add_struc_member(structure, sz_memname, offset, 0, NULL, size);
   save_struc(structure);               /* Darko, 02.05.2003 17:29 */
   //msg("adding %s at %d\n", sz_memname, offset);
#endif

   if (!op_stroff(ea, opnum, &(structure->id), 4, 0))
   {
      msg("[!] op_stroff failed !\n ea=%lx opnum = %lx\n", ea, opnum);
      return(false);
   }
   else
      return(true);
}
/* Darko, 29.04.2003 21:53 - end */

// returns the operand number in which "reg" is found
// If both operands use "reg", return the one which is not o_reg :)
int   __stdcall reg_opnd_num(int  reg, ea_t addr)
{
   ua_ana0(addr);

   if((cmd.Op1.reg == reg)&&(cmd.Op2.reg == reg))
   {
      if((cmd.Op1.type == o_reg)&&(cmd.Op2.type != o_imm)&&(cmd.Op2.type != o_mem))
         return(1);
      else
         return(0);
   }
   if(cmd.Op1.reg == reg)
      return(0);
   if(cmd.Op2.reg == reg)
      return(1);
   return(-1);
}


// This function will return the estimated number of arguments for a particular
// function
int   __stdcall get_func_arg_size(ea_t func)
{
   struc_t  *func_frame;
   member_t *ret_addr;
   int      arg_num;

   func_frame = get_frame(func);
   if(func_frame == 0)
      return(-1);
   ret_addr = get_member_by_name(func_frame, " r");
   if(ret_addr == 0)
      return(-1);

   arg_num = get_struc_last_offset(func_frame)-(ret_addr->soff);
   if(arg_num > 50)     // heuristic detection of skewed stackframes :)
      return(-1);
   return(arg_num);
}

#define MAX_STEPS 200
ea_t  __stdcall associate_push_to_call(ea_t ea_push)
{
   ea_t  next_call, func;
   int   arg_num, steps = MAX_STEPS;
   int   stack_push;
   func_t   *function;

   next_call = ea_push;
   ua_ana0(ea_push);

   function = get_func(ea_push);       // Get the function frame structure
   if((function == NULL)||(function == (void *)-1))
      return(BADADDR);                  /* Darko, 01.05.2003 23:00 */
   stack_push = get_spd(function, ea_push);

   while(steps-- != 0)     // step forwards
   {
      if(get_spd(function, next_call) < stack_push)
         break;
      while((cmd.itype != x86_callA)&&(cmd.itype != x86_callB)&&(cmd.itype != x86_callC))
      {
         if(cmd.itype == x86_retn)
            return(BADADDR);            /* Darko, 01.05.2003 23:01 */

         if(cmd.size == 0)
            next_call++;
         else
            next_call += cmd.size;
         ua_ana0(next_call);
      }  // trace until we find the next call

      func = get_next_cref_from(next_call, get_first_cref_from(next_call));

      if((func == 0)||(func == BADADDR))  /* Darko, 01.05.2003 22:56 */
         continue;

      ua_ana0(func); // now we need to deal with stub jmps
      if((cmd.itype == x86_jmpA)||(cmd.itype == x86_jmpB)||(cmd.itype == x86_jmpC)||(cmd.itype == x86_jmp_short))
         func = get_first_cref_from(func);

      if((func == BADADDR)||(func == 0))
         continue;

      arg_num = get_func_arg_size(func);
      if(arg_num == -1)
         continue;

      if((int)(get_spd(function, next_call)+arg_num) >= (stack_push)) /* [!] */
      {
         //msg("returning %lx\n", next_call);
         return(next_call);
      }
      else
         continue;
   }
   return(BADADDR);
}

ea_t  __stdcall register_xs(reg_t regnum, ea_t ea_begin, ea_t ea_end)
{
   while(ea_begin <= ea_end)
   {
      ua_ana0(ea_begin);

      if((cmd.Op1.type == o_reg)||(cmd.Op1.type == o_phrase)||(cmd.Op1.type == o_displ))
         if(cmd.Op1.reg == regnum)
            return(ea_begin);
      if((cmd.Op2.type == o_reg)||(cmd.Op2.type == o_phrase)||(cmd.Op2.type == o_displ))
         if(cmd.Op2.reg == regnum)
            return(ea_begin);
      if(cmd.size != 0)
         ea_begin += cmd.size;
      else
         ea_begin++;
   }
   if(ea_begin > ea_end)
      return(BADADDR);                  /* Darko, 01.05.2003 22:54 */
   return(ea_begin);
}
#endif
