#ifndef  __OBJREC
#define  __OBJREC
/*
   The contents of this file (and all files belonging to this project) are
   (c) Halvar Flake and Thomas Dullien

   Third attempt at writing a decent Structure/Object Reconstructor, this
   time with Code Flow/Code Block parsing.
   I should've learnt C++ way before - it seems to simplify a lot of things.

   Changelog:
         +  29th and 30th of August 2001
            Created graph.cpp, this file & ptrgraph.cpp with the intent
            of building a good Structure/Object reconstructor.
            Created an analogue to the flowgraphing facility first.
            Next step is to add pointer control blocks as graph nodes.
         +  31th of August 2001 & 1st of September
            Vacations are cool. Sun & Beach and one to three hours of
            coding a day seem to be quite productive.
            Added a (lame) graph sorting method to ptrgraph.cpp
            Added the entire pointer parsing crap to the Graph, seems
            to work fine.
         +  11th of September
            Coded a little more. Then the World Trade Center news came
            in. Quit coding.
         +  18th of September 2001
            Added a nice form for running the plugin. Wondering why I
            never used forms before - they are way nicer than the crap
            with going through multiple screens.
         +  20th of September 2001
            Started creating the flow graph code analysis parts.
         +  26th of September 2001
            Going to visit Belgium today. Trying to fix a few things in
            the code analyser while in the train.
            Urks. Never noticed how buggy the old code analyser is as I
            never had a working recursion. It sucks. Fixing bugs now...
            *heul* Ich will source-debugging !
            -  Bug 1:   Will improperly create struc member in case
                        current register is overwritten
            -  Bug 2:   Will improperly create struc member although
                        no access to the register is around (seems
                        to happen in 1-instruction blocks)
            -  Bug 3:   Fails to properly create struc member at "cmp"s
            But the code-flow module seems to have bugs as well:
            -  Bug 1:   Access violation at address 4 in complex graphs,
                        usually involving register recursion
        +   27th of September
            A day of bugfixing.
            Fixed most bugs in the analyser. Still not perfect, but doesn't
            create bogus members any more it seems. Lot's of nasty bugs came
            from the fact that I was using insn_t cmd directly and that got
            modified in-between.
            Nasty bug in the flowgrapher found - it doesn't seem to handle
            moving to/from stack variables to well yet. Looking into it.
            This bug seems to be the reason for so many structure members
            being created at 0xFFFFFxxx.
            Bug was in flow.cpp and comes from the fact that flow.cpp doesn't
            account for eax being destroyed in a call.
            Now on for the page fault <g> - basically a stupid problem. I forgot
            to check the return value of get_member_by_name in the pushcall-
            recursion.
        +   28th of September
            Fixed stupid bug that would create tons of bogus xrefs to the struc
            ture in analys.cpp
        +   30th of September
            Fixed another stupid bug and beautified the code a bit. Never ever
            use cmd, always make copies !
         +  1st of October
            Fixed bugs in the associate_push_to_all() to abort properly. Damned
            unpredictable compilers.
         +  8th of October
            Some more bugfixing, specifically in the analyser
            Changed stuff in legacy.cpp to support stuff like lea eax, [eax+XX],
            reg_opnd_num would screw up on it.
            Adding unintelligent, inheritance-ignorant vtable parsing
         +  23rd of February
            I am going to fix as many bugs as possible now and then release version
            1.0 (and discontinue this project) - A similar tool will be developed
            based on a meta-assebly in the future (if I find time)
            + fixed nasty bug in legacy.cpp - the code is bad:
               - 45: if(cmd.Op1.type == o_reg)
               + 45: if((cmd.Op1.type == o_reg)&&(cmd.Op2.type != o_imm)&&(cmd.Op2.type != o_mem))
              This is red-tape, but all of legacy.cpp seems to be incredibly badly
              written - fixing it from the ground up will probably break half the
              program.
              analys.cpp
               - 99: if(iOpNum == 0)
               + 99: if((iOpNum == 0)&&(cmd_A.Operands[iOpNum].type == o_reg))
              Should fix a few of the annoying directly negative structure members


*/

// IDA does a value->bool conversion, so disable the performance warning
#pragma warning( disable : 4800 )
// IDA references an internal type as both class and struct..
#pragma warning( disable : 4099 )


#include <ida.hpp>
#include <idp.hpp>
#include <loader.hpp>
#include <kernwin.hpp>
#include <frame.hpp>
#include <funcs.hpp>
#include <xref.hpp>
#include <bytes.hpp>
#include <nalt.hpp>
#include <ua.hpp>
#include <struct.hpp>
#include <bytes.hpp>
#include <typeinf.hpp>

#include "insx86.hpp"   // home-grown documentation of cmd.itype, actually
                        // there is a "real" version along from ilfak, in
                        // the future might even use it ;>
#include "objrec.hpp"
#include "ptrgraph.hpp" // pointer access graph class
#include "legacy.hpp"   // functions ported from strucrec v2
#include "flow.hpp"     // pointer flow graph functions
#include "analys.hpp"   // pointer analysis/code analysis


static reg_t RegToNum(char *szreg);

#if !defined(H_2_STROFF)
int __stdcall init(void)
{
   //msg("...initializing %s v%d.%d plugin...", PROGRAM_NAME, MAJOR_VERSION, MINOR_VERSION);

   if(strcmp(inf.procName, "metapc") != 0)
   {
      msg("  CPU Type should be \"metapc\" but is %s ! Aborting ...\n", inf.procName);
      return PLUGIN_SKIP;
   }
   //msg("  ... init successful !\n");
   return PLUGIN_OK;
}
#endif

//--------------------------------------------------------------------------
//      terminate

#if !defined(H_2_STROFF)
void __stdcall term(void)
{
}
#endif

//--------------------------------------------------------------------------
//       Convert a string to a reg_t
static reg_t RegToNum(char *szreg)
{
   const char  *name[]={"eax", "ecx", "edx", "ebx", "esp", "ebp", "esi", "edi"};
   unsigned int i;

   for(i = 0; i < 8; i++)
   {
      if(stricmp(szreg, name[i]) == 0)
         return((reg_t)i);
   }
   return(r_none);
}


#if !defined(H_2_STROFF)
//--------------------------------------------------------------------------
//      The plugin method
void __stdcall run(int arg)
{
   char  *regname, *strucname;
   // PtrGraph *grph;
   ea_t  curr;
   // member_t *stackvar;
   // reg_t    reg_ptr;
   rec_t    rec_mode;
   // unsigned int off;
   int      vtb_mode = 0;

   curr = get_screen_ea();
   regname = (char *)qalloc(40);
   strucname = (char *)qalloc(40);
   strcpy(strucname, "unkn_struc");
   strcpy(regname, "ecx");

   rec_mode = rec_full;

   AskUsingForm_c(cfg_box, &curr, strucname, regname, &rec_mode, &vtb_mode);

   if(vtb_mode)
   {
      while(get_first_dref_from(curr) != BADADDR)
      {
         ReconstructFromAddr(get_first_dref_from(curr), regname, rec_mode, strucname);
         curr += 4;
      }
   }
   else
      ReconstructFromAddr(curr, regname, rec_mode, strucname);

#ifdef DEBUG
   DbgFileCount = 0;
#endif
   qfree(regname);
   regname = 0;
   qfree(strucname);
   strucname = 0;
}
#endif                                  /* H_2_STROFF */


void  ReconstructFromAddr(ea_t curr, char *regname, rec_t rec_mode, char *strucname)
{
   member_t       *stackvar;
   reg_t          reg_ptr;
   unsigned int   off;
   PtrGraph       *grph;

   stackvar = get_member_by_name(get_frame(curr), regname);
   off = 0;
   if(!stackvar)
      reg_ptr = RegToNum(regname);
   else
   {
      reg_ptr = r_none;
      off = stackvar->soff;
   }

   grph = CreateFlowPtrGraph(curr, reg_ptr, off, rec_mode);
   strcpy(StrucName, strucname);

   grph->SortByBegin();
   grph->OutputWingraphFile("final.vcg");

   CreateStructFromGraph(grph, StrucName);
   delete grph;
}

#if !defined(H_2_STROFF)
//--------------------------------------------------------------------------
char comment[] = {PROGRAM_NAME "/metapc"};

char help[] =  HELP_STRING;
//--------------------------------------------------------------------------
// This is the preferred name of the plugin module in the menu system
// The preferred name may be overriden in plugins.cfg file

char wanted_name[] = {PROGRAM_NAME "/metapc"};

// This is the preferred hotkey for the plugin module
// The preferred hotkey may be overriden in plugins.cfg file
// Note: IDA won't tell you if the hotkey is not correct
//       It will just disable the hotkey.

char wanted_hotkey[] = WANTED_KEY;

//--------------------------------------------------------------------------
//
//      PLUGIN DESCRIPTION BLOCK
//
//--------------------------------------------------------------------------

extern "C" plugin_t PLUGIN = {
  IDP_INTERFACE_VERSION,
  0,                    // plugin flags
  init,                 // initialize
  term,                 // terminate. this pointer may be NULL.
  run,                  // invoke plugin
  comment,              // long comment about the plugin
                        // it could appear in the status line
                        // or as a hint
  help,                 // multiline help about the plugin
  wanted_name,          // the preferred short name of the plugin
  wanted_hotkey         // the preferred hotkey to run the plugin
};
#endif                                  /* H_2_STROFF */

#endif
